/*
 * @(#)Unzipper.java  1.1  2009-01-08
 *
 * Copyright (c) 2007-2009 Werner Randelshofer
 * Staldenmattweg 2, Immensee, CH-6405, Switzerland.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Werner Randelshofer. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Werner Randelshofer.
 */

package ch.randelshofer.velcrofly;

import com.ibm.icu.text.*;
import ch.randelshofer.gui.*;
import ch.randelshofer.util.*;
import java.io.*;
import java.util.zip.*;
import com.ice.tar.*;

/**
 * A Untarer uncompresses a tar archive into a directory.
 *
 * @author  Werner Randelshofer
 * @version 3.1 2009-01-08 If a file already exists, but does not have the
 * right case, delete it instead of just overwriting its contents.
 * <br>1.0  2007-11-19  Created.
 */
public class TarUnarchiver extends AbstractUnarchiver {
    
    /** Creates a new instance. */
    public TarUnarchiver(File archive) {
        super(archive, null);
    }
    /**
     * Creates a new instance.
     * @param archive The Zip archive.
     *
     * @param outputdir the output directory. If this value is null, the
     * output is written into the same directory which contains the archive.
     */
    public TarUnarchiver(File archive, File outputdir) {
        super(archive, archive);
    }
    
    /**
     * Unzips the ZIP archive from the provided input stream into
     * the specified output directory.
     * Closes input stream after completion.
     *
     * @param in An input stream containing a ZIP archive.
     * @param tgtdir The target directory.
     * @exception IOException when an I/O error occurs.
     */
    private void decompress(InputStream in, File tgtdir, ProgressView p)
    throws IOException {
        byte[] buf = new byte[512];
        TarInputStream zipin = null;
        OutputStream out = null;
        
        try {
            //if (! tgtdir.exists()) tgtdir.mkdirs();
            
            zipin = new TarInputStream(in);
            TarEntry entry = null;
            while ((entry = zipin.getNextEntry()) != null) {
                int pos = entry.getName().lastIndexOf("/");
                p.setNote("Extracting "+((pos == -1) ? entry.getName() : entry.getName().substring(pos+1)) );
                int max = p.getMaximum();
                p.setMaximum(max + 1);
                p.setProgress(max);
                
                if (entry.isDirectory()) {
                    File exportFile = toFile(tgtdir, entry);
                    exportFile.mkdirs();
                } else {
                    File exportFile = toFile(tgtdir, entry);

                    if (exportFile.exists() &&
                            exportFile.getCanonicalFile().getName().equals(exportFile.getName())) {
                        // If the filename does not match, we have to delete the
                        // file.
                        exportFile.delete();
                    } else {
                        if (!exportFile.getParentFile().exists()) {
                            exportFile.getParentFile().mkdirs();
                        }
                    }

                    try {
                        out = new FileOutputStream(exportFile);
                        int len;
                        while ((len = zipin.read(buf)) != -1) {
                            out.write(buf, 0, len);
                        }
                        out.close();
                        out = null;
                    } catch (FileNotFoundException e) {
                        String msg = "Unable to create a file named "+exportFile;
                        p.setWarning(msg);
                    }
                }
                //   zipin.closeEntry();
            }
        } finally {
            // We must not close the zip input stream here
            //if (zipin != null) try { zipin.close(); } catch (IOException e) {}
            if (out != null) try { out.close(); } catch (IOException e) {}
        }
    }
    
    public Object construct() {
        Object result = null;
        InputStream in = null;
        try {
            // Determine the target directory
            // and create it, if necessary.
            String name = Normalizer.normalize(archive.getName(), Normalizer.NFC);
            String parent = (outputdir == null) ? archive.getParent() : outputdir.getPath();
            boolean isGZipped = name.toLowerCase().endsWith(".tar.gz");
            File tgtdir;
            if (isGZipped) {
                name = name.substring(0, name.length() - 7);
            } else {
                int pos = name.lastIndexOf('.');
                if (pos != -1) {
                    name = name.substring(0, pos);
                }
            }
            tgtdir = new File(parent, name);
            int i = 1;
            while (tgtdir.exists()) {
                tgtdir = new File(parent, name+" "+(i++));
            }
            
            // decompress the archive into the target directory
            if (isGZipped) {
                decompress(in = new GZIPInputStream(new FileInputStream(archive)), tgtdir, p);
            } else {
                decompress(in = new FileInputStream(archive), tgtdir, p);
            }
            p.setNote("Completed.");
            p.close();
        } catch (Throwable e) {
            result = e;
            p.setNote(null);
            p.setError(""+e.getMessage());
            p.complete();
        } finally {
            if (in != null) {
                try { in.close(); } catch (IOException e) {if (result == null) result = e;}
            }
        }
        return result;
    }
    
    private File toFile(File targetDir, TarEntry entry) {
        String name = entry.getName();
        System.out.println(name);
        name = Normalizer.normalize(name, Normalizer.NFC);
        return new File(targetDir, Strings.translate(name, '/', File.separatorChar));
    }
    
    public String getArchiveName() {
        return "TAR";
    }
}
