/*
 * @(#)VelcroFly.java  3.1  2008-12-02
 *
 * Copyright (c) 2002-2008 Werner Randelshofer
 * Staldenmattweg 2, CH-6405 Immensee, Switzerland
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Werner Randelshofer. ("Confidential Information"). You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Werner Randelshofer.
 */
package ch.randelshofer.velcrofly;

import ch.randelshofer.gui.*;
import ch.randelshofer.io.Files;
import ch.randelshofer.util.*;

import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.io.*;
import java.util.*;
import java.util.prefs.Preferences;
import javax.swing.*;

/**
 * VelcroFly is a tiny zip compression and decompression tool.
 * To compress a file or a folder, drop it on VelcroFly's frame.
 * VelcroFly creates a zip file in the same folder of the dropped
 * file or folder.
 * To decompress a zip file, drop it on VelcoFly's frame.
 * If the zip archive contains only one file, it is decompressed in
 * the same directory where the zip archive was. If the zip archive
 * contains more than one file, VelcroFly creates a folder and
 * stores a file into the folder.
 *
 *
 * @author  Werner Randelshofer
 * @version 3.1 2008-12-02 Moved DnD-Code into an inner class named Handler.
 * <br>3.0 2007-11-19 Added support for unpacking TAR files. Added option
 * 'process each dropped file individually'.
 * <br>2.0 2005-01-06 Reworked.
 * <br>1.4.2 2004-03-26 Use best compression level to reduce
 * the size of zip archives.
 * <br>1.4.1 2003-10-31 Display an error message when decompression
 * fails. Files were not closed properly.
 * <br>1.4 2003-09-10 Decompress nested zip files recursively.
 * <br>1.3 2003-09-01 Sort files by filenames when creating nested zip.
 * This reduces the size of the archive considerably, when it contains multiple
 * copies of the same file.
 * <br>1.2 2003-05-12 Revised.
 * <br>1.1 2003-04-21 Revised.
 * <br>1.0.1 2002-12-16 The size of compressed files is wrong
 * after decompressing them.
 * <br>1.0 2002-11-20 Created.
 */
public class VelcroFly extends JFrame {

    private ConcurrentDispatcher dispatcher = new ConcurrentDispatcher();
    private Preferences pref;

    private class Handler implements DropTargetListener {

        /**
         * Called when a drag operation has
         * encountered the <code>DropTarget</code>.
         * <P>
         * @param dtde the <code>DropTargetDragEvent</code>
         */
        public void dragEnter(DropTargetDragEvent event) {
            if (event.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
                event.acceptDrag(DnDConstants.ACTION_COPY);
            } else {
                event.rejectDrag();
            }
        }

        /**
         * The drag operation has departed
         * the <code>DropTarget</code> without dropping.
         * <P>
         * @param dte the <code>DropTargetEvent</code>
         */
        public void dragExit(DropTargetEvent event) {
            // Nothing to do
        }

        /**
         * Called when a drag operation is ongoing
         * on the <code>DropTarget</code>.
         * <P>
         * @param dtde the <code>DropTargetDragEvent</code>
         */
        public void dragOver(DropTargetDragEvent event) {
            if (event.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
                event.acceptDrag(DnDConstants.ACTION_COPY);
            } else {
                event.rejectDrag();
            }
        }

        /**
         * The drag operation has terminated
         * with a drop on this <code>DropTarget</code>.
         * This method is responsible for undertaking
         * the transfer of the data associated with the
         * gesture. The <code>DropTargetDropEvent</code>
         * provides a means to obtain a <code>Transferable</code>
         * object that represents the data object(s) to
         * be transfered.<P>
         * From this method, the <code>DropTargetListener</code>
         * shall accept or reject the drop via the
         * acceptDrop(int dropAction) or rejectDrop() methods of the
         * <code>DropTargetDropEvent</code> parameter.
         * <P>
         * Subsequent to acceptDrop(), but not before,
         * <code>DropTargetDropEvent</code>'s getTransferable()
         * method may be invoked, and data transfer may be
         * performed via the returned <code>Transferable</code>'s
         * getTransferData() method.
         * <P>
         * At the completion of a drop, an implementation
         * of this method is required to signal the success/failure
         * of the drop by passing an appropriate
         * <code>boolean</code> to the <code>DropTargetDropEvent</code>'s
         * dropComplete(boolean success) method.
         * <P>
         * Note: The actual processing of the data transfer is not
         * required to finish before this method returns. It may be
         * deferred until later.
         * <P>
         * @param dtde the <code>DropTargetDropEvent</code>
         */
        public void drop(DropTargetDropEvent event) {
            if (event.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
                event.acceptDrop(DnDConstants.ACTION_COPY);

                try {
                    List<File> files = (List<File>) event.getTransferable().getTransferData(DataFlavor.javaFileListFlavor);
                    if (files.size() == 1 || processIndividuallyItem.isSelected()) {
                        for (File f : files) {
                            String lcName = f.getName().toLowerCase();
                            if (lcName.endsWith(".zip")) {
                                decompressZip(f);
                            } else if (lcName.endsWith(".tar") || lcName.endsWith(".tar.gz")) {
                                decompressTar(f);
                            } else {
                                LinkedList<File> compressList = new LinkedList<File>();
                                compressList.add(f);
                                compress(compressList);
                            }
                        }
                    } else {
                        compress(files);
                    }
                } catch (IOException e) {
                    JOptionPane.showConfirmDialog(
                            VelcroFly.this,
                            "Could not access the dropped data.",
                            "VelcroFly: Drop Failed",
                            JOptionPane.DEFAULT_OPTION, JOptionPane.INFORMATION_MESSAGE);
                } catch (UnsupportedFlavorException e) {
                    JOptionPane.showConfirmDialog(
                            VelcroFly.this,
                            "Unsupported data flavor.",
                            "VelcroFly: Drop Failed",
                            JOptionPane.DEFAULT_OPTION, JOptionPane.INFORMATION_MESSAGE);
                }
            } else {
                event.rejectDrop();
            }
        }

        /**
         * Called if the user has modified
         * the current drop gesture.
         * <P>
         * @param dtde the <code>DropTargetDragEvent</code>
         */
        public void dropActionChanged(DropTargetDragEvent event) {
            // Nothing to do
        }
    }
    private Handler handler;
    private LinkedList<JCheckBoxMenuItem> attributeItems;

    /** Creates new form VelcroFly */
    public VelcroFly() {
        initComponents();

        pref = Preferences.userNodeForPackage(VelcroFly.class);
        nestZipItem.setSelected(pref.getBoolean("nestZipFiles", false));
        processIndividuallyItem.setSelected(pref.getBoolean("processFilesIndividually", false));
        if (!Files.canWorkWithLabels()) {
            includeLabeledFilesSeparator.setVisible(false);
            includeUnlabeledItem.setVisible(false);
            includeRedLabelItem.setVisible(false);
            includeOrangeLabelItem.setVisible(false);
            includeYellowLabelItem.setVisible(false);
            includeGreenLabelItem.setVisible(false);
            includeBlueLabelItem.setVisible(false);
            includePurpleLabelItem.setVisible(false);
            includeGrayLabelItem.setVisible(false);

        } else {
            includeUnlabeledItem.setIcon(new FileLabelIcon(0));
            includeRedLabelItem.setIcon(new FileLabelIcon(6));
            includeOrangeLabelItem.setIcon(new FileLabelIcon(7));
            includeYellowLabelItem.setIcon(new FileLabelIcon(5));
            includeGreenLabelItem.setIcon(new FileLabelIcon(2));
            includeBlueLabelItem.setIcon(new FileLabelIcon(4));
            includePurpleLabelItem.setIcon(new FileLabelIcon(3));
            includeGrayLabelItem.setIcon(new FileLabelIcon(1));
        }
        attributeItems = new LinkedList<JCheckBoxMenuItem>();
        attributeItems.add(includeHiddenFilesItem);
        attributeItems.add(includeUnlabeledItem);
        attributeItems.add(includeRedLabelItem);
        attributeItems.add(includeOrangeLabelItem);
        attributeItems.add(includeYellowLabelItem);
        attributeItems.add(includeGreenLabelItem);
        attributeItems.add(includeBlueLabelItem);
        attributeItems.add(includePurpleLabelItem);
        attributeItems.add(includeGrayLabelItem);

        for (JCheckBoxMenuItem item : attributeItems) {
            item.setSelected(!pref.getBoolean(item.getActionCommand(), false));
        }

        handler = new Handler();
        new DropTarget(this, handler);
        new DropTarget(getContentPane(), handler);
        new DropTarget(label, handler);
    //nestZipCheckBoxMenuItem.setVisible(false);
    }

    public String getVersion() {
        String version = getClass().getPackage().getImplementationVersion();
        return (version == null) ? "unknown" : version;
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        label = new javax.swing.JLabel();
        menuBar = new javax.swing.JMenuBar();
        optionsMenu = new javax.swing.JMenu();
        nestZipItem = new javax.swing.JCheckBoxMenuItem();
        processIndividuallyItem = new javax.swing.JCheckBoxMenuItem();
        includeMenu = new javax.swing.JMenu();
        includeHiddenFilesItem = new javax.swing.JCheckBoxMenuItem();
        includeLabeledFilesSeparator = new javax.swing.JSeparator();
        includeUnlabeledItem = new javax.swing.JCheckBoxMenuItem();
        includeRedLabelItem = new javax.swing.JCheckBoxMenuItem();
        includeOrangeLabelItem = new javax.swing.JCheckBoxMenuItem();
        includeYellowLabelItem = new javax.swing.JCheckBoxMenuItem();
        includeGreenLabelItem = new javax.swing.JCheckBoxMenuItem();
        includeBlueLabelItem = new javax.swing.JCheckBoxMenuItem();
        includePurpleLabelItem = new javax.swing.JCheckBoxMenuItem();
        includeGrayLabelItem = new javax.swing.JCheckBoxMenuItem();
        viewMenu = new javax.swing.JMenu();
        statusMenuItem = new javax.swing.JMenuItem();
        helpMenu = new javax.swing.JMenu();
        aboutMenuItem = new javax.swing.JMenuItem();

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setTitle("VelcroFly");
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                exitForm(evt);
            }
        });

        label.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        label.setText("<html><font face=\"Dialog\">Drop a file on this window<br>to compress or decompress it.");
        label.setBorder(javax.swing.BorderFactory.createEmptyBorder(20, 20, 20, 20));
        getContentPane().add(label, java.awt.BorderLayout.CENTER);

        optionsMenu.setText("Options");

        nestZipItem.setText("Create Nested Zip");
        nestZipItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                nestZipPerformed(evt);
            }
        });
        optionsMenu.add(nestZipItem);

        processIndividuallyItem.setText("Process each dropped file individually");
        processIndividuallyItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                processIndividuallyPerformed(evt);
            }
        });
        optionsMenu.add(processIndividuallyItem);

        includeMenu.setText("Include Files with Attributes");

        includeHiddenFilesItem.setSelected(true);
        includeHiddenFilesItem.setText("Hidden");
        includeHiddenFilesItem.setActionCommand("skipHiddenFiles");
        includeHiddenFilesItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeHiddenFilesItem);
        includeMenu.add(includeLabeledFilesSeparator);

        includeUnlabeledItem.setSelected(true);
        includeUnlabeledItem.setText("No Label");
        includeUnlabeledItem.setActionCommand("skipUnlabeledFiles");
        includeUnlabeledItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeUnlabeledItem);

        includeRedLabelItem.setSelected(true);
        includeRedLabelItem.setText("Red Label");
        includeRedLabelItem.setActionCommand("skipRedFiles");
        includeRedLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeRedLabelItem);

        includeOrangeLabelItem.setSelected(true);
        includeOrangeLabelItem.setText("Orange Label");
        includeOrangeLabelItem.setActionCommand("skipOrangeFiles");
        includeOrangeLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeOrangeLabelItem);

        includeYellowLabelItem.setSelected(true);
        includeYellowLabelItem.setText("Yellow Label");
        includeYellowLabelItem.setActionCommand("skipYellowFiles");
        includeYellowLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeYellowLabelItem);

        includeGreenLabelItem.setSelected(true);
        includeGreenLabelItem.setText("Green Label");
        includeGreenLabelItem.setActionCommand("skipGreenFiles");
        includeGreenLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeGreenLabelItem);

        includeBlueLabelItem.setSelected(true);
        includeBlueLabelItem.setText("Blue Label");
        includeBlueLabelItem.setActionCommand("skipBlueFiles");
        includeBlueLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeBlueLabelItem);

        includePurpleLabelItem.setSelected(true);
        includePurpleLabelItem.setText("Purple Label");
        includePurpleLabelItem.setActionCommand("skipPurpleFiles");
        includePurpleLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includePurpleLabelItem);

        includeGrayLabelItem.setSelected(true);
        includeGrayLabelItem.setText("Gray Label");
        includeGrayLabelItem.setActionCommand("skipGrayFiles");
        includeGrayLabelItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                includeAttributedFilePerformed(evt);
            }
        });
        includeMenu.add(includeGrayLabelItem);

        optionsMenu.add(includeMenu);

        menuBar.add(optionsMenu);

        viewMenu.setText("View");

        statusMenuItem.setText("Status");
        statusMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                showStatus(evt);
            }
        });
        viewMenu.add(statusMenuItem);

        menuBar.add(viewMenu);

        helpMenu.setText("Help");

        aboutMenuItem.setText("About");
        aboutMenuItem.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                about(evt);
            }
        });
        helpMenu.add(aboutMenuItem);

        menuBar.add(helpMenu);

        setJMenuBar(menuBar);

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void processIndividuallyPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_processIndividuallyPerformed
        pref.putBoolean("processFilesIndividually", processIndividuallyItem.isSelected());
    }//GEN-LAST:event_processIndividuallyPerformed

    private void nestZipPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_nestZipPerformed
        pref.putBoolean("nestZipFiles", nestZipItem.isSelected());
    }//GEN-LAST:event_nestZipPerformed

    private void showStatus(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_showStatus
        // TODO add your handling code here:
        ProgressFrame.getInstance().show();
    }//GEN-LAST:event_showStatus

    private void about(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_about
        JOptionPane.showConfirmDialog(
                this,
                "<html>" + Fonts.emphasizedDialogFontTag("VelcroFly") + Fonts.smallDialogFontTag("<br>" +
                "Version " + getVersion() + "<br><br>" +
                "Copyright \u00a9 2003-2008<br>" +
                "Werner Randelshofer<br>" +
                "Staldenmattweg 2, CH-6405 Immensee<br>" +
                "Switzerland<br>" +
                "mailto:werner.randelshofer@bluewin.ch<br>" +
                "http://www.randelshofer.ch/<br>" +
                "All rights reserved.<br><br>" +
                "com.ice.tar library by<br>" +
                "Timothy Gerard Endres<br>" +
                "mailto:time@gjt.org<br>" +
                "Public domain.<br><br>"+
                "Java "+System.getProperty("java.version")+"<br>"+
                System.getProperty("os.name")+" "+System.getProperty("os.version")+" "+System.getProperty("os.arch")),
                "About VelcroFly",
                JOptionPane.DEFAULT_OPTION, JOptionPane.PLAIN_MESSAGE);

    }//GEN-LAST:event_about

    /** Exit the Application */
    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm
        System.exit(0);
    }//GEN-LAST:event_exitForm

    private void includeAttributedFilePerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_includeAttributedFilePerformed
        JCheckBoxMenuItem item = (JCheckBoxMenuItem) evt.getSource();
        pref.putBoolean(item.getActionCommand(), !item.isSelected());
    }//GEN-LAST:event_includeAttributedFilePerformed

    public void setEnabled(boolean b) {
        super.setEnabled(b);
        menuBar.setEnabled(b);
        optionsMenu.setEnabled(b);
        includeHiddenFilesItem.setEnabled(b);
        nestZipItem.setEnabled(b);
    }

    /**
     * Decompresses the zip archive.
     * If the zip archive contains only one file or folder,
     * it is written into the same directory as the archive.
     * If the zip archive contains more than one file or folder,
     * then they are written into a folder. The name of
     * the folder is the same as the name of the zip archive
     * with the suffix '.zip' removed.
     */
    public void decompressZip(File archive) {
        dispatcher.dispatch(new ZipUnarchiver(archive, pref.getBoolean("skipHiddenFiles", true)));
    }

    /**
     * Decompresses the zip archive.
     * If the zip archive contains only one file or folder,
     * it is written into the same directory as the archive.
     * If the zip archive contains more than one file or folder,
     * then they are written into a folder. The name of
     * the folder is the same as the name of the zip archive
     * with the suffix '.zip' removed.
     */
    public void decompressTar(File archive) {
        dispatcher.dispatch(new TarUnarchiver(archive));
    }

    private FileFilter createFileFilter() {
//                includeHiddenFilesItem.isSelected(),
//                (includeRedLabelItem.isSelected()) ? 6 : -1));

        final boolean skipHidden = pref.getBoolean("skipHiddenFiles", false);
        final int skipLabels;
        if (Files.canWorkWithLabels()) {
            skipLabels =
                    (pref.getBoolean("skipUnlabeledFiles", false) ? 1<<0 : 0) |
                    (pref.getBoolean("skipRedFiles", false) ? 1<<6 : 0) |
                    (pref.getBoolean("skipOrangeFiles", false) ? 1<<7 : 0) |
                    (pref.getBoolean("skipYellowFiles", false) ? 1<<5 : 0) |
                    (pref.getBoolean("skipGreenFiles", false) ? 1<<2 : 0) |
                    (pref.getBoolean("skipBlueFiles", false) ? 1<<4 : 0) |
                    (pref.getBoolean("skipPurpleFiles", false) ? 1<<3 : 0) |
                    (pref.getBoolean("skipGrayFiles", false) ? 1<<1 : 0);
        } else {
            skipLabels = 0;
        }

        System.out.println("VelcroFly skipLabels:"+skipLabels);

        return new FileFilter() {
            public boolean accept(File file) {
                if (skipHidden && file.isHidden()) {
                    return false;
                }
                if (skipLabels != 0) {
                    int labelBit = 1 << Files.getLabel(file);
                    if ((skipLabels & labelBit) != 0) {
                        return false;
                    }
                }
                return true;
            }
        };
    }

    /**
     * Creates a zip archive and recursively stores all
     * files and folders in it.
     * The name of the zip archive is generated by appending
     * ".zip" to the name of the first file.
     */
    private void compressAll(final List filesAndDirs) {
        dispatcher.dispatch(new ZipArchiver(
                filesAndDirs,
                null,
                createFileFilter(),
                nestZipItem.isSelected()));
    }

    /**
     * Creates a zip archive and recursively stores all
     * files and folders in it.
     * The name of the zip archive is generated by appending
     * ".zip" to the name of the first file.
     */
    public void compress(List filesAndDirs) {
        Iterator i = filesAndDirs.iterator();
        String parent = ((File) filesAndDirs.get(0)).getParent();
        boolean allFilesHaveSameParent = true;
        while (i.hasNext()) {
            File f = (File) i.next();
            if (!f.getParent().equals(parent)) {
                allFilesHaveSameParent = false;
                break;
            }
        }

        if (allFilesHaveSameParent) {
            compressAll(filesAndDirs);
        } else {
            i = filesAndDirs.iterator();
            while (i.hasNext()) {
                ArrayList l = new ArrayList();
                l.add(i.next());
                compressAll(l);
            }
        }
    }
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuItem aboutMenuItem;
    private javax.swing.JMenu helpMenu;
    private javax.swing.JCheckBoxMenuItem includeBlueLabelItem;
    private javax.swing.JCheckBoxMenuItem includeGrayLabelItem;
    private javax.swing.JCheckBoxMenuItem includeGreenLabelItem;
    private javax.swing.JCheckBoxMenuItem includeHiddenFilesItem;
    private javax.swing.JSeparator includeLabeledFilesSeparator;
    private javax.swing.JMenu includeMenu;
    private javax.swing.JCheckBoxMenuItem includeOrangeLabelItem;
    private javax.swing.JCheckBoxMenuItem includePurpleLabelItem;
    private javax.swing.JCheckBoxMenuItem includeRedLabelItem;
    private javax.swing.JCheckBoxMenuItem includeUnlabeledItem;
    private javax.swing.JCheckBoxMenuItem includeYellowLabelItem;
    private javax.swing.JLabel label;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JCheckBoxMenuItem nestZipItem;
    private javax.swing.JMenu optionsMenu;
    private javax.swing.JCheckBoxMenuItem processIndividuallyItem;
    private javax.swing.JMenuItem statusMenuItem;
    private javax.swing.JMenu viewMenu;
    // End of variables declaration//GEN-END:variables
}
