/*
 * @(#)Zipper.java  3.0  2008-12-04
 *
 * Copyright (c) 2004-2008 Werner Randelshofer
 * Staldenmattweg 2, Immensee, CH-6405, Switzerland.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Werner Randelshofer. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Werner Randelshofer.
 */
package ch.randelshofer.velcrofly;

import com.ibm.icu.text.*;
import ch.randelshofer.gui.*;
import ch.randelshofer.io.Files;
import ch.randelshofer.util.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import javax.swing.*;

/**
 * A Zipper compresses a list of files and directories into a zip archive.
 * The Zipper can create a nested zip archive for better compression.
 *
 * @author  Werner Randelshofer
 * @version 3.0 2008-12-04 Factored file filtering out into a FileFilter.
 * <br>2.0.1 2008-12-04 Fixed a minor bug with the file separator char
 * in addZipEntry.
 * <br>2.0 2008-04-05 Added skip color.
 * <br>1.1 2006-10-09 Normalize filenames to Unicode normal form NFC 
 * before creating ZipEntry objects.
1.0  06 January 2005  Created.
 */
public class ZipArchiver extends RunnableWorker {

    /**
     * List of files and directories.
     */
    private List filesAndDirs;
    /**
     * Zip Archive.
     */
    private File archive;
    /**
     * Progress view.
     */
    private ProgressView p;
    private boolean isNested;
    private FileFilter filter;

    /**
     * Creates a new instance.
     *
     * @param filesAndDirs A list of files and directories. List items must be
     * instances of class java.io.List. All files and directories must have the
     * same parent directory!
     */
    public ZipArchiver(List filesAndDirs) {
        this(filesAndDirs, null, null, false);
    }

    /**
     * Creates a new instance.
     *
     * @param filesAndDirs A list of files and directories. List items must be
     * instances of class java.io.List. All files and directories must have the
     * same parent directory!
     * @param archive The name of the zip archive. If the name is null, the
     * name of the archive is derived from the first file name.
     * @param isNested Set this to true, to create a nested zip archive.
     * @param isSkipHiddenFiles Set this to true, to skip hidden files.
     * @param skipLabel Set this to a value between 0 and 7 to skip files
     * with this label. Set this to -1 to not skip labeled files.
     */
    public ZipArchiver(List filesAndDirs, File archive, FileFilter filter, boolean isNested) {
        this.filesAndDirs = filesAndDirs;
        this.filter = (filter != null) ? filter : new FileFilter() {

            public boolean accept(File pathname) {
                return true;
            }
        };
        this.isNested = isNested;

        if (archive == null) {
            File f = (File) filesAndDirs.get(0);
            archive = new File(f.getParent(), f.getName() + ((isNested) ? ".nested.zip" : ".zip"));
            for (int i = 2; i < Integer.MAX_VALUE && archive.exists(); i++) {
                archive = new File(f.getParent(), f.getName() + "." + i + ((isNested) ? ".nested.zip" : ".zip"));
            }
        }
        this.archive = archive;

        p = new ProgressView(
                ((isNested) ? "Creating Nested ZIP Archive \"" : "Creating ZIP Archive \"") + archive.getName() + "\"",
                "Waiting for processor...");
    }

    private void listRecursive(List l, File f) {
        if (filter.accept(f)) {
            l.add(f);

            if (f.isDirectory()) {
                File[] children = f.listFiles();
                for (int i = 0; i < children.length; i++) {
                    listRecursive(l, children[i]);
                }
            }
        }
    }

    private void addZipEntry(ZipOutputStream out, String path, File f)
            throws IOException {

        // Create a suitable file name for the zip archive
        String name;
        if (f.getPath().startsWith(path) && !f.getPath().equals(path)) {
            name = f.getPath().substring(path.length() + 1);
        } else {
            name = f.getPath();
        }
        name = name.replace(File.separatorChar, '/');
        name = Normalizer.normalize(name, Normalizer.NFC);
        if (f.isDirectory()) {
            // f is a directory.
            // only create an entry, if the directory is empty.
            if (f.list().length == 0) {

                // Ensure that the name ends with a '/' character.
                // Create an entry for the directory
                if (name.length() == 0 || name.charAt(name.length() - 1) != '/') {
                    name += File.separatorChar;
                }
                ZipEntry entry = new ZipEntry(name);
                out.putNextEntry(entry);
                out.closeEntry();
            }
        } else {
            // f is a file:
            // Create the zip entry and
            // copy the contents of the file into it
            ZipEntry entry = new ZipEntry(name);
            out.putNextEntry(entry);

            //entry.setSize(f.length());
            FileInputStream in = new FileInputStream(f);
            try {
                byte[] buf = new byte[512];
                int len;
                while ((len = in.read(buf)) != -1) {
                    out.write(buf, 0, len);
                }
            } finally {
                in.close();
            }
            out.closeEntry();
        }
    }

    public Object construct() {

        Object result = Boolean.FALSE;
        ZipOutputStream out = null;
        ZipOutputStream zipout = null;

        try {
            // Determine the files that need to be compressed
            p.setNote("Determining Files to be added...");
            p.setIndeterminate(true);

            List files = new LinkedList();
            Iterator i = filesAndDirs.iterator();
            while (i.hasNext()) {
                File f = (File) i.next();
                listRecursive(files, f);
            }

            p.setMaximum(files.size());
            p.setIndeterminate(false);
            p.setNote("Files to add: " + files.size());

            if (files.size() > 65535) {
                throw new IOException("Too many files: " + files.size());
            }

            if (isNested) {
                // The idea of a nested zip is to get a better compression
                // because we have lots of similar files.
                // We put all files together of which we think they are
                // similar.
                // Files with the same filename extension and with the
                // same name are likely to be similar.
                p.setNote("Sorting files...");
                Collections.sort(files, new Comparator() {

                    public int compare(Object o1, Object o2) {
                        File f1 = (File) o1;
                        File f2 = (File) o2;
                        String n1 = f1.getName();
                        String n2 = f2.getName();

                        // Compare by filename extension first
                        int result = 0;
                        int p1 = n1.lastIndexOf('.');
                        int p2 = n2.lastIndexOf('.');
                        if (p1 == -1 && p2 != -1) {
                            result = -1;
                        } else if (p1 != -1 && p2 == -1) {
                            result = 1;
                        } else if (p1 != -1 && p2 != -1) {
                            result = n1.substring(p1).compareTo(n2.substring(p2));
                        }
                        // Then Compare by name
                        if (result == 0) {
                            result = n1.compareTo(n2);
                        }
                        return result;
                    }
                });
            }
            // Create the zip archive
            String path = ((File) filesAndDirs.get(0)).getParent();

            zipout = out = new ZipOutputStream(new FileOutputStream(archive));
            out.setLevel(Deflater.BEST_COMPRESSION);

            if (isNested) {
                String name = ((File) filesAndDirs.get(0)).getName() + ".zip";
                name = Normalizer.normalize(name, Normalizer.NFC);
                out.putNextEntry(new ZipEntry(name));
                out = new ZipOutputStream(out);
                out.setLevel(Deflater.NO_COMPRESSION);
            }

            int count = 0;
            i = files.iterator();
            while (i.hasNext() && !p.isCanceled()) {
                p.setProgress(count++);

                File entry = (File) i.next();

                p.setNote("Adding " + count + "/" + files.size() + " " + entry.getName());
                addZipEntry(out, path, entry);
            }
            result = (p.isCanceled()) ? Boolean.FALSE : Boolean.TRUE;
            p.setNote("Completed.");
            p.close();
        } catch (Throwable e) {
            result = e;
            p.setError("" + e.getMessage());
            p.complete();
        } finally {
            if (out != null) {
                try {
                    out.close();
                } catch (IOException e) {
                    if (result != Boolean.FALSE) {
                        result = e;
                    }
                }
            }
            if (zipout != null && zipout != out) {
                try {
                    zipout.close();
                } catch (IOException e) {
                    if (result != Boolean.FALSE) {
                        result = e;
                    }
                }
            }
            if (result == Boolean.FALSE && archive.exists()) {
                archive.delete();
            }
        }

        return result;
    }

    /**
     * Called on the event dispatching thread (not on the worker thread)
     * after the <code>construct</code> method has returned.
     */
    public void finished(Object result) {
        /*
        if (result == Boolean.FALSE) {
        JOptionPane.showConfirmDialog(
        null,
        "Creation of ZIP Archive \""+archive.getName()+"\" canceled.",
        "VelcroFly",
        JOptionPane.DEFAULT_OPTION, JOptionPane.PLAIN_MESSAGE
        );
        } else
         */
        if (result != null && (result instanceof Throwable)) {
            ((Throwable) result).printStackTrace();
            JOptionPane.showConfirmDialog(
                    null,
                    "Creation of ZIP Archive \"" + archive.getName() + "\" failed.\n" + ((Throwable) result).getMessage(),
                    "VelcroFly",
                    JOptionPane.DEFAULT_OPTION, JOptionPane.ERROR_MESSAGE);
        }
    }
}
