/*
 * @(#)ZipUnarchiver.java  3.1  2009-01-08
 *
 * Copyright (c) 2004-2009 Werner Randelshofer
 * Staldenmattweg 2, Immensee, CH-6405, Switzerland.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information of
 * Werner Randelshofer. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Werner Randelshofer.
 */
package ch.randelshofer.velcrofly;

import com.ibm.icu.text.*;
import ch.randelshofer.gui.*;
import ch.randelshofer.util.*;
import java.io.*;
import java.util.zip.*;

/**
 * Uncompresses a zip archive into a directory.
 *
 * @author  Werner Randelshofer
 * @version 3.1 2009-01-08 If a file already exists, but does not have the
 * right case, delete it instead of just overwriting its contents.
 * <br>3.0 2008-04-05 Added support for skipping hidden files.
 * <br>2.0 2007-11-19 Renamed from Unzipper to ZipUnarchiver, moved
 * reusable code into abstract base class.
 * <br>1.1.1 2007-11-12 Skip files which can't be unzipped due to
 * naming conventions on the operating system we are currently running on.
 * <br>1.1 2006-10-09 Normalize filenames to Unicode normal form NFC
 * before creating file objects.
 * <br>1.0  06 January 2005  Created.
 */
public class ZipUnarchiver extends AbstractUnarchiver {

    private boolean skipHiddenFiles;

    /** Creates a new instance. */
    public ZipUnarchiver(File archive, boolean skipHiddenFiles) {
        super(archive, null);
        this.skipHiddenFiles = skipHiddenFiles;
    }

    /**
     * Creates a new instance.
     * @param archive The Zip archive.
     *
     * @param outputdir the output directory. If this value is null, the
     * output is written into the same directory which contains the archive.
     */
    public ZipUnarchiver(File archive, File outputdir, boolean skipHiddenFiles) {
        super(archive, outputdir);
        this.skipHiddenFiles = skipHiddenFiles;
    }

    /**
     * Unzips the ZIP archive from the provided input stream into
     * the specified output directory.
     * Closes input stream after completion.
     *
     * @param in An input stream containing a ZIP archive.
     * @param tgtdir The target directory.
     * @param recursive If this is true, zip files whithin the archive will
     * be unzipped.
     * @param skipHiddenFiles If this is true, files which start with '.' are
     * not exported.
     * @exception IOException when an I/O error occurs.
     */
    private void decompress(InputStream in, File tgtdir, boolean recursive, ProgressView p)
            throws IOException {
        byte[] buf = new byte[512];
        ZipInputStream zipin = null;
        OutputStream out = null;

        try {
            zipin = new ZipInputStream(in);
            ZipEntry entry = null;
            while ((entry = zipin.getNextEntry()) != null) {
                int pos = entry.getName().lastIndexOf("/");

                String entryFileName = ((pos == -1) ? entry.getName() : entry.getName().substring(pos + 1));

                if (skipHiddenFiles &&
                        (entryFileName.startsWith(".") || entry.getName().indexOf("/.") != -1)) {
                    p.setNote("Skipping " + entryFileName);
                    continue;
                }

                p.setNote("Extracting " + entryFileName);

                int max = p.getMaximum();
                p.setMaximum(max + 1);
                p.setProgress(max);

                if (entry.isDirectory()) {
                    File exportFile = toFile(tgtdir, entry);
                    exportFile.mkdirs();
                } else if (recursive && entry.getName().endsWith(".zip")) {
                    File exportFile = toFile(tgtdir.getParentFile(), entry);
                    exportFile = new File(exportFile.getPath().substring(0, exportFile.getPath().length() - 4));
                    if (exportFile.exists() || exportFile.isDirectory()) {
                        decompress(zipin, tgtdir, false, p);
                    } else {
                        decompress(zipin, tgtdir.getParentFile(), false, p);
                    }
                } else {
                    File exportFile = toFile(tgtdir, entry);

                    if (exportFile.exists() &&
                            exportFile.getCanonicalFile().getName().equals(exportFile.getName())) {
                        // If the filename does not match, we have to delete the
                        // file.
                        exportFile.delete();
                    } else {
                        if (!exportFile.getParentFile().exists()) {
                            exportFile.getParentFile().mkdirs();
                        }
                    }

                    try {
                        out = new FileOutputStream(exportFile);
                        int len;
                        while ((len = zipin.read(buf)) != -1) {
                            out.write(buf, 0, len);
                        }
                        out.close();
                        out = null;
                    } catch (FileNotFoundException e) {
                        String msg = "Unable to create a file named " + exportFile;
                        p.setWarning(msg);
                    }
                }
                zipin.closeEntry();
            }
        } finally {
            // We must not close the zip input stream here
            //if (zipin != null) try { zipin.close(); } catch (IOException e) {}
            if (out != null) {
                try {
                    out.close();
                } catch (IOException e) {
                }
            }
        }
    }

    public Object construct() {
        Object result = null;
        InputStream in = null;
        try {
            // Determine the target directory
            // and create it, if necessary.
            String name = Normalizer.normalize(archive.getName(), Normalizer.NFC);
            String parent = (outputdir == null) ? archive.getParent() : outputdir.getPath();
            boolean isNested = name.toLowerCase().endsWith(".nested.zip");
            File tgtdir;
            if (isNested) {
                name = name.substring(0, name.length() - 11);
            } else {
                int pos = name.lastIndexOf('.');
                if (pos != -1) {
                    name = name.substring(0, pos);
                }
            }
            tgtdir = new File(parent, name);
            int i = 1;
            while (tgtdir.exists()) {
                tgtdir = new File(parent, name + " " + (i++));
            }

            // decompress the archive into the target directory
            decompress(in = new FileInputStream(archive), tgtdir, isNested, p);
            p.setNote("Completed.");
            p.close();
        } catch (Throwable e) {
            result = e;
            p.setNote(null);
            p.setError("" + e.getMessage());
            p.complete();
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException e) {
                    if (result == null) {
                        result = e;
                    }
                }
            }
        }
        return result;
    }

    private File toFile(File targetDir, ZipEntry entry) {
        String name = entry.getName();
        name = Normalizer.normalize(name, Normalizer.NFC);
        return new File(targetDir, Strings.translate(name, '/', File.separatorChar));
    }

    public String getArchiveName() {
        return "ZIP";
    }
}
